package org.dataone.tidy.merge;

import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.TreeSet;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.dataone.cn.dao.exceptions.DataAccessException;
import org.dataone.configuration.Settings;
import org.dataone.service.exceptions.ServiceFailure;
import org.dataone.service.types.TypeCompareUtil;
import org.dataone.service.types.tidy.TidyAccessMap;
import org.dataone.service.types.tidy.TidyReplicaList;
import org.dataone.service.types.v1.Identifier;
import org.dataone.service.types.v1.NodeReference;
import org.dataone.service.types.v1.SystemMetadata;
import org.dataone.tidy.concurrent.AbstractTidyJob;
import org.dataone.tidy.concurrent.TidyJob;
import org.dataone.tidy.dao.ChangeRecord;
import org.dataone.tidy.dao.MergeStatus;
import org.dataone.tidy.merge.strategy.MergeSystemMetadata;

/**
 * This class contains the execution logic for merging one pid, given the appropriate
 * input.  It is responsible also for filling out the change records portion
 * of the report, and persisting results.
 * 
 * 
 * @author rnahf
 *
 */
public class MergeJob extends AbstractTidyJob implements TidyJob {

    private static Log logger = LogFactory.getLog(MergeJob.class);

    private static final String ORC_NODE_ID = Settings.getConfiguration().getString("tidy.dao.orc.nodeid");
    private static final String UCSB_NODE_ID = Settings.getConfiguration().getString("tidy.dao.ucsb.nodeid");
    private static final String UNM_NODE_ID = Settings.getConfiguration().getString("tidy.dao.unm.nodeid");


    public MergeJob(Identifier pid) {
        this.pid = pid;
    }

    public MergeJob() {
    }

    /**
     * This call method interacts with the data sources and repositories to
     * execute the merge strategy, and will record either Success or Failure
     * in the MergeResults repository.  It will only throw an exception if
     * it cannot save to the MergeResults repository.
     * 
     */
    @Override
    public Boolean call() throws Exception {
        logger.info("Running MergeJob for " + getPid().getValue());
        Boolean returnValue = true;
        Map<String, SystemMetadata> cnSysmetaMap;
        MergeStrategy<Map<String, SystemMetadata>> mergeStrategy;

        MergeReport report = new MergeReport(this.pid);

        try {
            cnSysmetaMap = fetchCNSysmetas();
            if (cnSysmetaMap != null && !(cnSysmetaMap.isEmpty())) {
            	mergeStrategy = new MergeSystemMetadata(report, this.getMnReadClient());
                mergeStrategy.mend(cnSysmetaMap);
                buildSysmetaChangesReport(cnSysmetaMap, report);
            
            } else {
                logger.info("No CN systemmetadata to work with for pid: " + this.pid);
                report.setReview("No CN systemmetadata: probable delete");
            }
            

            
        } catch (DataAccessException daox) {
            report.setProcessingException(daox);

        } catch (RuntimeException rte) {
            report.setProcessingException(rte);

        } finally {

            // make sure there is no merged sysmeta to record if there's been an exception
            if (report.getProcessingException() != null) {
                report.setResultingSystemMetadata(null);
                this.setFailure(report.getProcessingException());
                returnValue = false;
            }

            try {
                persistMergeReport(report);
            } catch (Exception e) {
                logger.error(e.getMessage(), e);
                this.setFailure(e);
                returnValue = false;
            }
        }
        return returnValue;
    }

    /** 
     *  Only creating an entry in the sysmeta map if we have a sysmeta entry 
     *  this will help processing by the mergeStrategy
     * @return
     * @throws DataAccessException
     */
    protected Map<String, SystemMetadata> fetchCNSysmetas() throws DataAccessException {

        Map<NodeReference, SystemMetadata> cnSysMetas = sysmetaRepo.getSystemMetadata(getPid());
        if (cnSysMetas == null || cnSysMetas.isEmpty()) {
            logger.error("fetchCNSysmetas: sysmetaRepo.getSystemMetadata (pid = " + 
            		getPid().getValue() + ") returned an empty or null map (" + cnSysMetas + ")");
        }
        Map<String, SystemMetadata> sysmetas = new HashMap<String, SystemMetadata>();
        if (cnSysMetas != null) {
            for (NodeReference nr : cnSysMetas.keySet()) {
                if (cnSysMetas.get(nr) != null) {
                    sysmetas.put(nr.getValue(), cnSysMetas.get(nr));
                } else {
                    logger.warn(nr.getValue() + " in fetchCNSysmetas cnSysMetas " +
                    		"returned a null value (pid = " + getPid().getValue() + ")");
                }
            }
        }
        return sysmetas;
    }

    /**
     * attempts to persist the elements of the report to their respective
     * repositories. If the merge result (summary of the execution) cannot 
     * be persisted, then callers should expect a runtime exception  
     * 
     * @param report
     */
    protected void persistMergeReport(MergeReport report) {

        if (report.getProcessingException() == null) {
            try {
            	logger.info("*****Change Records [repo.Save in] (pid:" + report.getIdentifier().getValue() +
                		"): " + report.getChangeRecordList().size());
                Iterable<ChangeRecord> cr = changesRepo.save(report.getChangeRecordList());
                logger.info("*****Change Records [repo.Save out] (pid:" + report.getIdentifier().getValue() +
                		"): " + countChangeRecordsReturned(cr));
                
                if (report.getResultingSystemMetadata() != null) {
                	sysmetaRepo.saveSystemMetadata(report.getResultingSystemMetadata());
                }
                report.getMergeResult().setMergeStatus(MergeStatus.SUCCESS);
            } catch (Exception e) {
                report.setProcessingException(e);
                e.printStackTrace();
                report.getMergeResult().setMergeStatus(MergeStatus.FAILURE);
                report.setReview(null);
            }
        } else {
            report.getMergeResult().setMergeStatus(MergeStatus.FAILURE);
            report.setReview(null);
        }

        // a report is always saved, an if it can't be saved a runtime exception
        // will be thrown 
        resultsRepo.save(report.getMergeResult());
    }
    
    
    private Integer countChangeRecordsReturned(Iterable<ChangeRecord> cr) {
    	
    	if (cr == null) return null;
    	
    	if (cr instanceof Collection) {
    		return ((Collection<?>) cr).size();
    	}

    	int count = 0;
    	for (ChangeRecord i : cr) {
    		count++;
    	}
    	return count;
    }

    /**
     * adapted from TypeCompareUtils in d1_common_java to produce output that
     * could be written directly into the ChangeRecord object
     * 
     * @param sysmetaMap
     * @param mergeReport
     * @param sortArrays
     */
    public void buildSysmetaChangesReport(Map<String, SystemMetadata> sysmetaMap,
            MergeReport mergeReport) {

        // need to include the authmn and resulting sysmetas into the comparison
        if (mergeReport.getAuthMnSystemMetadata() != null) {
            SystemMetadata mnSysmeta = mergeReport.getAuthMnSystemMetadata();
            sysmetaMap.put("AuthMN", mnSysmeta);
        }

        if (mergeReport.getResultingSystemMetadata() != null) {
            SystemMetadata resultingSysmeta = mergeReport.getResultingSystemMetadata();
            sysmetaMap.put("resulting", resultingSysmeta);
        }


        // derive subtype listings for all (potentailly) 5 sysmetas
        Map<String, Map<String,String>> d1TypePropMap = new HashMap<String, Map<String,String>>();
        
        d1TypePropMap.put(ORC_NODE_ID, buildSubtypeListing(sysmetaMap.get(ORC_NODE_ID)));
        d1TypePropMap.put(UCSB_NODE_ID, buildSubtypeListing(sysmetaMap.get(UCSB_NODE_ID)));
        d1TypePropMap.put(UNM_NODE_ID, buildSubtypeListing(sysmetaMap.get(UNM_NODE_ID)));
        d1TypePropMap.put("AuthMN", buildSubtypeListing(sysmetaMap.get("AuthMN")));
        d1TypePropMap.put("resulting", buildSubtypeListing(sysmetaMap.get("resulting")));
        
        
        // create an uber-keyset for comparison time
        Set<String> allAttributes = new TreeSet<String>();
        for (String label: d1TypePropMap.keySet()) {
        	allAttributes.addAll(d1TypePropMap.get(label).keySet());
        }

        for (String attribute : allAttributes) {
        	ChangeRecord cr = deriveChangeRecord(mergeReport.getIdentifier(), attribute, d1TypePropMap );
        	if (cr != null) {
        		mergeReport.addChangeRecord(cr);
        	}
        }
    }
     

    private ChangeRecord deriveChangeRecord(Identifier pid, String attribute, 
    		Map<String,Map<String,String>> subtypeListingsMap)
    {
    	ChangeRecord changeRecord = new ChangeRecord(pid.getValue());
    	changeRecord.setAttribute(attribute);

    	// ignore the listings for the AccessPolicy-proper - we only want
    	// to report from the custom tidy/MaxPermissionMap
    	// (pardon the hackiness)
    	if (attribute.contains("AccessPolicy") && !attribute.contains("tidy")) {
    		return null;
    	} else if (attribute.contains("ReplicaList") && !attribute.contains("tidy")) {
    		return null;
    	}

    	logger.debug("buildChangeReport for '" + pid.getValue() + "' check property: " + attribute);

    	List<String> distinctValues = new LinkedList<String>();
    	Set<String> nodeLabels = subtypeListingsMap.keySet();

    	for (String label : nodeLabels) {

    		// don't contribute any disinctValues if the sysmetadata is not available
    		// (unless it's the //systemMetadata line)
    		if ((subtypeListingsMap.get(label).get("//SystemMetadata").equals("Not Available"))
    				&& (attribute.equals("//SystemMetadata") == false)) {
    			continue;
    		}

    		String value = ((LinkedHashMap<String, String>) subtypeListingsMap.get(label)).get(attribute);

    		if (label.equals(ORC_NODE_ID)) {
    			changeRecord.setCnORC(value);
    		} else if (label.equals(UNM_NODE_ID)) {
    			changeRecord.setCnUNM(value);
    		} else if (label.equals(UCSB_NODE_ID)) {
    			changeRecord.setCnUCSB(value);
    		} else if (label.equals("AuthMN")) {
    			changeRecord.setAuthMN(value);
    		} else if (label.equals("resulting")) {
    			changeRecord.setResulting(value);
    		}

    		// collect unique values
    		// massaging unique values in some cases

    		// a null value for archived is equivalent to 'false'
    		if (attribute.contains("Archived") && (value == null || value.equals("null"))) {
    			value = Boolean.FALSE.toString();
    		}

    		// sometimes we get empty strings as values, which are 
    		// equivalent to null, so make it so...
    		if (value == null) value = "null";
    		if (value.equals("")) value = "null"; 



    		if (!distinctValues.contains(value) && label.equals("AuthMN") == false) {
    			distinctValues.add(value);
    		} else if (!distinctValues.contains(value) && label.equals("AuthMN")
    				&& attribute.contains("AccessPolicy")) {
    			distinctValues.add(value);
    		}
    	}
    	if (distinctValues.size() > 1) {
    		return changeRecord;
    	}
    	return null;

    }
    
    
    
    private Map<String,String> buildSubtypeListing(SystemMetadata smd) {      
        
    	Map<String,String> subtypesMap;
    	
    	if (smd != null) {
    		
    	
    		// AccessPolicy differences are difficult to read in the ChangeRecord
    		// so using a custom service type that transforms the AccessPolicy into
    		// a map structure that should be easier to scan as a ChangeRecord

    		// create the TidyAccessMap and remove the AccessPolicy from the sysmeta
    		TidyAccessMap tam = new TidyAccessMap(smd.getAccessPolicy());

    		TidyReplicaList tidyRepls = new TidyReplicaList(smd.getReplicaList());

    		// get subtype listing for the accesspolicy-less sysmeta
    		subtypesMap = TypeCompareUtil.getD1SubtypesListing("",smd, true);

    		// get subtype listing from the new access map
    		Map<String,String> tamProps = TypeCompareUtil.getD1SubtypesListing("/AccessPolicy/tidy", tam, false);

    		// get subtype listing from the new (compressed) replica list
    		Map<String,String> tidyReplProps = TypeCompareUtil.getD1SubtypesListing("/ReplicaList/tidy", tidyRepls, true);


    		// add the access map subtype listing to the sysmeta one
    		subtypesMap.putAll(tamProps);
    		subtypesMap.putAll(tidyReplProps);
    		subtypesMap.put("//SystemMetadata","Present");
    	
    	} else {
    		subtypesMap = new LinkedHashMap<String,String>();
    		subtypesMap.put("//SystemMetadata","Not Available");
    	}

        // shove the reassembled listings into the propMap
        return subtypesMap;
    }

}
